# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2._helpers import (
    opener,
    saver,
    toc,
    pdf_renderer,
    page_renderer,
    text_inserter,
    text_extractor,
)
from pypdfium2 import _pypdfium as pdfium


class PdfDocument:
    """
    Document helper class.
    
    Parameters:
        input_obj: The file or data to load.
        password: A password to unlock the PDF, if encrypted.
    """
    
    def __init__(
            self,
            input_obj,
            password = None,
        ):
        self._input_obj = input_obj
        self._password = password
        self._pdf, self._ld_data = opener.open_pdf_auto(
            self._input_obj,
            password = self._password,
        )
    
    def __len__(self):
        """ Get the number of pages the PDF has. """
        return pdfium.FPDF_GetPageCount(self._pdf)
    
    @property
    def raw(self):
        """ Get the raw PDFium ``FPDF_DOCUMENT`` handle. """
        return self._pdf
    
    def close(self):
        """ Close the document to release allocated memory. This method must be called when done with processing the PDF. """
        return opener.close_pdf(self._pdf, self._ld_data)
    
    def save(self, *args, **kws):
        """ Save the PDF into an output byte buffer (see :func:`.save_pdf`). """
        return saver.save_pdf(self._pdf, *args, **kws)
    
    def get_page(self, index):
        """ Get the page at *index* (``FPDF_PAGE``). """
        return opener.open_page(self._pdf, index)
    
    def get_textpage(self, index):
        """ Get the text page at *index* (:class:`.PdfTextPage`). """
        return text_extractor.PdfTextPage(self.get_page(index), close_input=True)
    
    def get_toc(self, **kws):
        """ Incrementally read the document's table of contents (see :func:`.get_toc`). """
        yield from toc.get_toc(self._pdf, **kws)
    
    def render_page_tobytes(self, index, **kws):
        """ Render a single page to bytes (see :func:`.render_page_tobytes`). """
        return page_renderer.render_page_tobytes(self._pdf, page_index=index, **kws)
    
    def render_pdf_tobytes(self, **kws):
        """ Incrementally render multiple pages to bytes (see :func:`.render_pdf_tobytes`). """
        yield from pdf_renderer.render_pdf_tobytes(self._input_obj, password=self._password, **kws)
    
    def render_page_topil(self, index, **kws):
        """ Render a single page to a :mod:`PIL` image (see :func:`.render_page_topil`). """
        return page_renderer.render_page_topil(self._pdf, page_index=index, **kws)
    
    def render_pdf_topil(self, **kws):
        """ Incrementally render multiple pages to :mod:`PIL` images (see :func:`.render_pdf_topil`). """
        yield from pdf_renderer.render_pdf_topil(self._input_obj, password=self._password, **kws)
    
    def insert_text(self, index, **kws):
        """ Insert text into a page (see :func:`.insert_text`). """
        return text_inserter.insert_text(self._pdf, page_index=index, **kws)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.nativeopener import (
    is_buffer,
    open_pdf_buffer,
)
from pypdfium2._helpers.error_handler import handle_pdfium_error


def open_pdf_auto(input_obj, password=None):
    """    
    Open a document from a file path or in-memory data.
    
    Parameters:
        input_obj (str | bytes | typing.BinaryIO):
            File path to a PDF document, bytes, or a byte buffer.
        password (typing.Optional[str | bytes]):
            A password to unlock the document, if encrypted.
    
    Returns:
        ``Tuple[pdfium.FPDF_DOCUMENT, Optional[LoaderDataHolder]]`` –
        The handle to a PDFium document, and optionally a loader data object to store associated buffer data.
    
    Warning:
        Callers **MUST** ensure that the loader data object remain available for as long as they work with the PDF. That means it has to be accessed again when done with processing, to prevent Python from automatically deleting the object. This can be achieved by passing it as second parameter to :func:`.close_pdf`, which is also necessary to release the acquired file buffer. If attempting to access the ``FPDF_DOCUMENT`` handle after the loader data has been deleted, a segmentation fault would occur.
    """
    
    ld_data = None
    if isinstance(input_obj, bytes):
        input_obj = io.BytesIO(input_obj)
    
    if isinstance(input_obj, str):
        # file path -> use FPDF_LoadDocument()
        pdf = pdfium.FPDF_LoadDocument(input_obj, password)
        if pdfium.FPDF_GetPageCount(pdf) < 1:
            handle_pdfium_error(False)
    elif is_buffer(input_obj):
        # byte buffer -> use open_pdf_buffer()
        pdf, ld_data = open_pdf_buffer(input_obj, password)
    else:
        raise ValueError("Input must be a file path, bytes or a byte buffer, but it is %s." % type(input_obj))
    
    return pdf, ld_data


def close_pdf(pdf, loader_data=None):
    """
    Close an in-memory PDFium document.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            The PDFium document object to close using ``FPDF_CloseDocument()``.
        loader_data (typing.Optional[LoaderDataHolder]):
            Object that stores custom file access data associated to the PDF.
    """
    
    pdfium.FPDF_CloseDocument(pdf)
    if loader_data is not None:
        loader_data.close()


class PdfContext:
    """
    Context manager to open and automatically close again a PDFium document.
    
    Parameters:
        input_obj: The file or data to load.
        password: A password to unlock the PDF, if encrypted.
    
    Returns:
        ``FPDF_DOCUMENT`` handle to the raw PDFium object.
    """
    
    def __init__(
            self,
            input_obj,
            password = None,
        ):
        self.input_obj = input_obj
        self.password = password
        self.ld_data = None
    
    def __enter__(self):
        self.pdf, self.ld_data = open_pdf_auto(
            self.input_obj,
            password = self.password,
        )
        return self.pdf
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        close_pdf(self.pdf, self.ld_data)


def open_page(pdf, page_index):
    """
    Get a handle to the page at a certain index. When you have finished working with the page, call :func:`.close_page`.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            PDFium document handle.
        page_index (int):
            Zero-based index of the page to load.
    
    Returns:
        ``FPDF_PAGE``
    """
    
    page_count = pdfium.FPDF_GetPageCount(pdf)
    if not 0 <= page_index < page_count:
        raise IndexError("Page index %s is out of bounds for document with %s pages." % (page_index, page_count))
    
    return pdfium.FPDF_LoadPage(pdf, page_index)


def close_page(page):
    """
    Close a PDFium page object (``FPDF_PAGE``) to release allocated memory.
    This is a wrapper around ``FPDF_ClosePage()``.
    """
    return pdfium.FPDF_ClosePage(page)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import logging
from os.path import join, dirname, abspath

lib_logger = logging.getLogger('pypdfium2')
lib_logger.addHandler(logging.StreamHandler())

TestDir     = dirname(abspath(__file__))
SourceTree  = dirname(TestDir)
ResourceDir = join(TestDir,'resources')
OutputDir   = join(TestDir,'output')

sys.path.insert(0, join(SourceTree,'setupsrc'))
from pl_setup.packaging_base import PlatformNames


class TestFiles:
    render             = join(ResourceDir,'render.pdf')
    encrypted          = join(ResourceDir,'encrypted.pdf')
    multipage          = join(ResourceDir,'multipage.pdf')
    bookmarks          = join(ResourceDir,'bookmarks.pdf')
    bookmarks_circular = join(ResourceDir,'bookmarks_circular.pdf')
    boxes              = join(ResourceDir,'boxes.pdf')
    text               = join(ResourceDir,'text.pdf')
    empty              = join(ResourceDir,'empty.pdf')
    images             = join(ResourceDir,'images.pdf')


def iterate_testfiles(skip_encrypted=True):
    encrypted = (TestFiles.encrypted, )
    for attr_name in dir(TestFiles):
        if attr_name.startswith('_'):
            continue
        member = getattr(TestFiles, attr_name)
        if skip_encrypted and member in encrypted:
            continue
        yield member


def get_members(cls):
    members = []
    for attr in dir(cls):
        if attr.startswith('_'):
            continue
        members.append(attr)
    return members


pl_names = get_members(PlatformNames)



# SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from os.path import join
from importlib.util import find_spec
import pytest
import pypdfium2 as pdfium
from ..conftest import ResourceDir, OutputDir


@pytest.mark.skipif(not find_spec("uharfbuzz"), reason="uharfbuzz is not installed")
def test_insert_text():
    
    pdf = pdfium.FPDF_CreateNewDocument()
    pdfium.FPDFPage_New(pdf, 0, 500, 800)
    
    NotoSans = join(ResourceDir, "NotoSans-Regular.ttf")
    font_info = pdfium.FontInfo(NotoSans)
    pdf_font = pdfium.open_pdffont(
        pdf, NotoSans,
        type = pdfium.FPDF_FONT_TRUETYPE,
        is_cid = True,
    )
    
    pdfium.insert_text(
        pdf, 0,
        text = "मैं घोषणा, पुष्टि और सहमत हूँ कि:",
        pos_x = 100,
        pos_y = 700,
        font_size = 25,
        font_info = font_info,
        pdf_font = pdf_font,
    )
    pdfium.insert_text(
        pdf, 0,
        text = "Latin letters test.",
        pos_x = 100,
        pos_y = 600,
        font_size = 30,
        font_info = font_info,
        pdf_font = pdf_font,
    )
    
    pdfium.close_pdffont(pdf_font)
    
    with open(join(OutputDir, "text_insertion.pdf"), "wb") as buffer:
        pdfium.save_pdf(pdf, buffer, version=17)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
from pypdfium2 import _pypdfium as pdfium


def get_pageobjs(page):
    """
    Get all page objects present on the given page.
    
    Yields:
        ``FPDF_PAGEOBJECT``
    """
    n_objects = pdfium.FPDFPage_CountObjects(page)
    for i in range(n_objects):
        yield pdfium.FPDFPage_GetObject(page, i)

def filter_pageobjs(page_objects, obj_type):
    """
    Get itmes of a certain type from a sequence of page objects.
    
    Parameters:
        obj_type (``FPDF_PAGEOBJ_...``):
            The object type to filter.
    Yields:
        ``FPDF_PAGEOBJECT``
    """
    for page_obj in page_objects:
        if pdfium.FPDFPageObj_GetType(page_obj) == obj_type:
            yield page_obj

def locate_pageobj(page_object):
    """
    Get the bounding box of a page object.
    
    Returns:
        Coordinates for (left, bottom, right, top).
    """
    left, bottom, right, top = [ctypes.c_float() for _ in range(4)]
    ret_code = pdfium.FPDFPageObj_GetBounds(page_object, *[ctypes.byref(coord) for coord in (left, bottom, right, top)])
    if not ret_code:
        raise RuntimeError("Locating the page object failed")
    return [coord.value for coord in (left, bottom, right, top)]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import pypdfium2 as pdfium


def test_translate_rotation():
    test_cases = [
        (0,   0),
        (90,  1),
        (180, 2),
        (270, 3),
    ]
    for input, expectation in test_cases:
        assert pdfium.translate_rotation(input) == expectation


def test_colour_to_hex():
    test_cases = [
        ((255, 255, 255),      (0xFFFFFFFF, False)),
        ((255, 255, 255, 255), (0xFFFFFFFF, False)),
        ((0,   255, 255, 255), (0xFF00FFFF, False)),
        ((255, 0,   255, 255), (0xFFFF00FF, False)),
        ((255, 255, 0,   255), (0xFFFFFF00, False)),
        ((255, 255, 255, 0  ), (0x00FFFFFF, True )),
    ]
    for values, expectation in test_cases:
        assert pdfium.colour_as_hex(*values) == expectation



# SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.opener import open_page

try:
    import uharfbuzz as harfbuzz
except ImportError:
    have_harfbuzz = False
else:
    have_harfbuzz = True


class FontInfo:
    """
    Object to cache harfbuzz font data.
    
    Parameters:
        font_path (str):
            File path of the font to use.
    """
    
    def __init__(self, font_path):
        if not have_harfbuzz:
            raise RuntimeError("Font helpers require uharfbuzz to be installed.")
        self.hb_blob = harfbuzz.Blob.from_file_path(font_path)
        self.hb_face = harfbuzz.Face(self.hb_blob)
        self.hb_font = harfbuzz.Font(self.hb_face)
        self.hb_scale = self.hb_font.scale[0]


def open_pdffont(pdf, font_path, type, is_cid):
    """
    Create a PDFium font object handle. When you have finished working with the font, call :func:`.close_pdffont`.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            PDFium document handle.
        font_path (str):
            File path of the font to use.
        type (int):
            Type of the given font (``FPDF_FONT_TYPE1`` or ``FPDF_FONT_TRUETYPE``).
        is_cid (bool):
            :data:`True` if the font is a CID font, :data:`False` otherwise.
    Returns:
        ``FPDF_FONT``
    """
    with open(font_path, "rb") as fh:
        data = fh.read()
    return pdfium.FPDFText_LoadFont(
        pdf,
        ctypes.cast(data, ctypes.POINTER(ctypes.c_uint8)),
        len(data),
        type,
        is_cid,
    )


def close_pdffont(pdf_font):
    """
    Close a PDFium font object handle.
    
    Parameters:
        pdf_font (``FPDF_FONT``):
            The font object to close.
    """
    return pdfium.FPDFFont_Close(pdf_font)


def insert_text(
        pdf,
        page_index,
        text,
        pos_x,
        pos_y,
        font_size,
        font_info,
        pdf_font,
    ):
    """
    Insert text into a PDF page at a specified position using the writing system's ligature.
    This function supports Asian scripts such as Hindi. ``uharfbuzz`` is required as an additional dependency.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            PDFium document handle.
        page_index (int):
            Zero-based index of the page on which to add text.
        text (str):
            The message to insert.
        pos_x (float):
            Distance from left border to first character.
        pos_y (float):
            Distance from bottom border to first character.
        font_size (float):
            Font size the text shall have.
        font_info (FontInfo):
            Font data and information.
        pdf_font (``FPDF_FONT``):
            PDFium font object handle.
    """
    
    page = open_page(pdf, page_index)
    
    hb_buffer = harfbuzz.Buffer()
    hb_buffer.add_str(text)
    hb_buffer.guess_segment_properties()
    hb_features = {"kern": True, "liga": True}
    harfbuzz.shape(font_info.hb_font, hb_buffer, hb_features)
    
    start_point = pos_x
    for info, pos in zip(hb_buffer.glyph_infos, hb_buffer.glyph_positions):
        pdf_textobj = pdfium.FPDFPageObj_CreateTextObj(pdf, pdf_font, font_size)
        pdfium.FPDFText_SetCharcodes(pdf_textobj, ctypes.c_uint32(info.codepoint), 1)
        pdfium.FPDFPageObj_Transform(
            pdf_textobj,
            1, 0, 0, 1,
            start_point - (pos.x_offset / font_info.hb_scale) * font_size,
            pos_y,
        )
        pdfium.FPDFPage_InsertObject(page, pdf_textobj)
        start_point += (pos.x_advance / font_info.hb_scale) * font_size
    
    pdfium.FPDFPage_GenerateContent(page)
    pdfium.FPDF_ClosePage(page)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
from ..conftest import TestFiles
import pypdfium2 as pdfium


def test_pageobjects():
    doc = pdfium.PdfDocument(TestFiles.images)
    page = doc.get_page(0)
    
    pageobjs = pdfium.get_pageobjs(page)
    images = [img for img in pdfium.filter_pageobjs(pageobjs, pdfium.FPDF_PAGEOBJ_IMAGE)]
    assert len(images) == 3
    
    positions = [pdfium.locate_pageobj(img) for img in images]
    exp_positions = [
        [133, 459, 350, 550],
        [48, 652, 163, 700],
        [204, 204, 577, 360],
    ]
    assert len(positions) == len(exp_positions)
    for pos, exp_pos in zip(positions, exp_positions):
        assert pytest.approx(pos, abs=1) == exp_pos
    
    pdfium.close_page(page)
    doc.close()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import shutil
from os.path import dirname, abspath, join

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import SourceTree, run_cmd


Git = shutil.which('git')
RepositoryURL = 'https://github.com/pypdfium2-team/pypdfium2'


def run_default(command):
    return run_cmd(command, cwd=SourceTree, capture=True)

def get_url(tag):
    return RepositoryURL + '/tree/%s' % tag

def generate_tags_list():
    tags_str = run_default([Git, 'for-each-ref', '--sort=creatordate', '--format', '%(refname)', 'refs/tags'])
    tags_list = tags_str.split('\n')
    tags_list.reverse()
    return tags_list


def get_tag(TagsList, n_descends, skip_beta=False):
    
    i = 0
    for line in TagsList:
        tag = line.split('/')[-1].strip()
        if skip_beta and i > 0 and 'b' in tag:
            continue
        if i >= n_descends:
            break
        i += 1
    
    return tag


def main():
    
    TagsList = generate_tags_list()
    current_tag, prev_tag = get_tag(TagsList, 0), get_tag(TagsList, 1, True)
    print(current_tag, prev_tag)
    
    relnotes = "Release %s\n\n" % current_tag
    relnotes += "### Changes\n\nCommits between [`%s`](%s) and [`%s`](%s):\n\n" % (prev_tag, get_url(prev_tag), current_tag, get_url(current_tag))
    relnotes += run_default([Git, 'log', '%s..%s' % (prev_tag, current_tag), "--pretty=format:* %H %s"])
    relnotes += '\n'
    
    with open(join(SourceTree, 'RELEASE.md'), 'w') as fh:
        fh.write(relnotes)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
import logging
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.utilities import translate_viewmode

logger = logging.getLogger(__name__)


class OutlineItem:
    """
    An entry in the table of contents ("bookmark").
    
    Parameters:
        level (int):
            The number of parent items.
        title (str):
            String of the bookmark.
        is_closed (bool):
            Whether child items are hidden by default.
        page_index (int):
            Zero-based index of the page the bookmark is pointing to.
        view_mode (ViewMode):
            A mode defining how to interpret the coordinates of *view_pos*.
        view_pos (typing.Sequence[float]):
            Target position on the page the viewport should jump to. It is a sequence of float values in PDF points. Depending on *view_mode*, it can contain between 0 and 4 coordinates.
    """
    
    def __init__(
            self,
            level,
            title,
            is_closed,
            page_index,
            view_mode,
            view_pos,
        ):
        self.level = level
        self.title = title
        self.is_closed = is_closed
        self.page_index = page_index
        self.view_mode = view_mode
        self.view_pos = view_pos


def _get_toc_entry(pdf, bookmark, level):
    """ Convert a raw PDFium bookmark to an :class:`.OutlineItem`. """
    
    # title
    t_buflen = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
    t_buffer = ctypes.create_string_buffer(t_buflen)
    pdfium.FPDFBookmark_GetTitle(bookmark, t_buffer, t_buflen)
    title = t_buffer.raw[:t_buflen].decode('utf-16-le')[:-1]
    
    # page index
    dest = pdfium.FPDFBookmark_GetDest(pdf, bookmark)
    page_index = pdfium.FPDFDest_GetDestPageIndex(pdf, dest)
    
    # state
    is_closed = pdfium.FPDFBookmark_GetCount(bookmark) < 0
    
    # viewport
    n_params = ctypes.c_ulong()
    view_pos = (pdfium.FS_FLOAT * 4)()
    view_mode = pdfium.FPDFDest_GetView(dest, n_params, view_pos)
    view_pos = list(view_pos)[:n_params.value]
    view_mode = translate_viewmode(view_mode)
    
    return OutlineItem(
        level = level,
        title = title,
        is_closed = is_closed,
        page_index = page_index,
        view_mode = view_mode,
        view_pos = view_pos,
    )


def get_toc(
        pdf,
        parent = None,
        level = 0,
        max_depth = 15,
        seen = None,
    ):
    """
    Parse the outline ("table of contents") of a PDF document.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            The PDFium document of which to read the outline.
        max_depth (int):
            The maximum recursion depth to consider when analysing the table of contents.
    Yields:
        :class:`OutlineItem`
    """
    
    if level >= max_depth:
        return []
    if seen is None:
        seen = set()
    
    bookmark = pdfium.FPDFBookmark_GetFirstChild(pdf, parent)
    
    while bookmark:
        
        address = ctypes.addressof(bookmark.contents)
        if address in seen:
            logger.critical("A circular bookmark reference was detected whilst parsing the table of contents.")
            break
        else:
            seen.add(address)
        
        yield _get_toc_entry(pdf, bookmark, level)
        yield from get_toc(
            pdf,
            parent = bookmark,
            level = level + 1,
            max_depth = max_depth,
            seen = seen,
        )
        
        bookmark = pdfium.FPDFBookmark_GetNextSibling(pdf, bookmark)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
import os.path
from enum import Enum
from pypdfium2 import _namespace as pdfium


class Units (Enum):
    PT = 0
    MM = 1
    CM = 2
    IN = 3


def units_to_pt(value, unit: Units):
    if unit is Units.PT:
        return value
    elif unit is Units.IN:
        return value*72
    elif unit is Units.CM:
        return (value*72) / 2.54
    elif unit is Units.MM:
        return (value*72) / 25.4
    else:
        raise ValueError("Invalid unit type %s" % unit)


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        'tile',
        help = "Perform page tiling (N-up compositing)",
    )
    parser.add_argument(
        'input',
        help = "PDF file on which to perform N-up compositing",
    )
    parser.add_argument(
        '--output', '-o',
        required = True,
        type = os.path.abspath,
        help = "Target path for the new document",
    )
    parser.add_argument(
        '--rows', '-r',
        type = int,
        required = True,
        help = "Number of rows (horizontal tiles)",
    )
    parser.add_argument(
        '--cols', '-c',
        type = int,
        required = True,
        help = "Number of columns (vertical tiles)",
    )
    parser.add_argument(
        '--width',
        type = float,
        required = True,
        help = "Target width",
    )
    parser.add_argument(
        '--height',
        type = float,
        required = True,
        help = "Target height",
    )
    parser.add_argument(
        '--unit', '-u',
        default = Units.MM,
        type = lambda string: Units[string.upper()],
        help = "Unit for target width and height (pt, mm, cm, in)",
    )    


def main(args):
    
    width = units_to_pt(args.width, args.unit)
    height = units_to_pt(args.height, args.unit)
    
    with pdfium.PdfContext(args.input) as src_pdf:
        
        dest_pdf = pdfium.FPDF_ImportNPagesToOne(
            src_pdf,
            ctypes.c_float(width),       # output_width
            ctypes.c_float(height),      # output_height
            ctypes.c_size_t(args.cols),  # num_pages_on_x_axis
            ctypes.c_size_t(args.rows),  # num_pages_on_y_axis
        )
        
        with open(args.output, 'wb') as file_handle:
            pdfium.save_pdf(dest_pdf, file_handle)
        
        pdfium.close_pdf(dest_pdf)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import shutil
import tempfile
from os.path import join
import pypdfium2 as pdfium
import pytest
from ..conftest import TestFiles


def _load_pdfct(file_or_data, password=None, page_count=1):
    with pdfium.PdfContext(file_or_data, password) as pdf:
        assert isinstance(pdf, pdfium.FPDF_DOCUMENT)
        assert pdfium.FPDF_GetPageCount(pdf) == page_count


def test_pdfct_str():
    in_path = TestFiles.render
    assert isinstance(in_path, str)
    _load_pdfct(in_path)


def test_pdfct_bytes():
    with open(TestFiles.render, 'rb') as file:
        data = file.read()
        assert isinstance(data, bytes)
    _load_pdfct(data)


def test_pdfct_bytesio():
    with open(TestFiles.render, 'rb') as file:
        buffer = io.BytesIO(file.read())
        assert isinstance(buffer, io.BytesIO)
    _load_pdfct(buffer)
    buffer.close()


def test_pdfct_bufreader():
    with open(TestFiles.render, 'rb') as buf_reader:
        assert isinstance(buf_reader, io.BufferedReader)
        _load_pdfct(buf_reader)


def test_pdfct_encrypted():
    _load_pdfct(TestFiles.encrypted, 'test_user')
    _load_pdfct(TestFiles.encrypted, 'test_owner')
    _load_pdfct(TestFiles.encrypted, 'test_user')
    with open(TestFiles.encrypted, 'rb') as buf_reader:
        _load_pdfct(buf_reader, password='test_user')


def test_pdfct_encrypted_fail():
    pw_err_context = pytest.raises(pdfium.PdfiumError, match="Missing or wrong password.")
    with pw_err_context:
        _load_pdfct(TestFiles.encrypted)
    with pw_err_context:
        _load_pdfct(TestFiles.encrypted, 'string')
    with pw_err_context:
        _load_pdfct(TestFiles.encrypted, 'string')


def test_open_native():
    
    pdf, ld_data = pdfium.open_pdf_native(TestFiles.multipage)
    
    pdfium.FPDFPage_Delete(pdf, 1)
    page = pdfium.FPDF_LoadPage(pdf, 0)
    rotation = pdfium.FPDFPage_GetRotation(page)
    pdfium.FPDF_ClosePage(page)
    
    pdfium.close_pdf(pdf, ld_data)


def test_open_nonascii_pdfct():
    with tempfile.TemporaryDirectory(prefix="pypdfium2_") as tempdir:
        nonascii_file = join(tempdir,"nonascii_tênfilechứakýtựéèáàçß 发短信.pdf")
        shutil.copy(TestFiles.render, nonascii_file)
        _load_pdfct(nonascii_file)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0

# Configuration file for the Sphinx documentation builder.
# See https://www.sphinx-doc.org/en/master/usage/configuration.html

import time

project = 'pypdfium2'
author = 'pypdfium2-team'
copyright = '%s %s' % (time.strftime('%Y'), author)

html_theme = 'sphinx_rtd_theme'
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinxcontrib.programoutput',
    'myst_parser',
]

autodoc_preserve_defaults = True
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'member-order': 'bysource',
}
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'PIL': ('https://pillow.readthedocs.io/en/stable/', None),
}



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _namespace as pdfium


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        'toc',
        help = "Show a PDF document's table of contents",
    )
    parser.add_argument(
        'input',
        help = "PDF document of which to print the outline",
    )
    parser.add_argument(
        '--max-depth',
        type = int,
        default = 15,
        help = "Maximum recursion depth to consider when parsing the table of contents",
    )


def print_toc(toc):
    for item in toc:
        print(
            '    ' * item.level +
            '[{}] '.format('-' if item.is_closed else '+') +
            '{} -> {}  # {} {}'.format(
                item.title,
                item.page_index + 1,
                item.view_mode,
                item.view_pos,
            )
        )


def main(args):
    doc = pdfium.PdfDocument(args.input)
    print_toc( doc.get_toc() )
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import ctypes
from pypdfium2 import _pypdfium as pdfium


class _writer_class:
    
    def __init__(self, buffer):
        self.buffer = buffer
        if not callable( getattr(self.buffer, 'write', None) ):
            raise ValueError("Output buffer must implement the write() method.")
    
    def __call__(self, _filewrite, data, size):
        block = ctypes.cast(data, ctypes.POINTER(ctypes.c_ubyte * size))
        self.buffer.write(block.contents)
        return 1


def save_pdf(pdf, buffer, version=None):
    """
    Write the data of a PDFium document into an output buffer.
    
    Parameters:
        pdf (``FPDF_DOCUMENT``):
            Handle to a PDFium document.
        buffer (typing.BinaryIO):
            A byte buffer to capture the data. It may be anything that implements the ``write()`` method.
        version (typing.Optional[int]):
            The PDF version to use, given as an integer (14 for 1.4, 15 for 1.5, ...).
            If :data:`None`, PDFium will set a version automatically.
    """
    
    WriteFunctype = ctypes.CFUNCTYPE(
        # restype
        ctypes.c_int,
        # argtypes
        ctypes.POINTER(pdfium.FPDF_FILEWRITE),
        ctypes.POINTER(None),
        ctypes.c_ulong,
    )
    
    filewrite = pdfium.FPDF_FILEWRITE()
    filewrite.WriteBlock = WriteFunctype( _writer_class(buffer) )
    
    saveargs = (pdf, ctypes.byref(filewrite), pdfium.FPDF_NO_INCREMENTAL)
    if version is None:
        pdfium.FPDF_SaveAsCopy(*saveargs)
    else:
        pdfium.FPDF_SaveWithVersion(*saveargs, version)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import math
import ctypes
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.opener import open_page
from pypdfium2._helpers.constants import OptimiseMode
from pypdfium2._helpers.utilities import (
    colour_as_hex,
    translate_rotation,
)

try:
    import PIL.Image
except ImportError:
    have_pil = False
else:
    have_pil = True


def _get_clformat(use_alpha, greyscale):
    
    px = 'BGRA', pdfium.FPDFBitmap_BGRA
    
    if not use_alpha:
        if greyscale:
            px = 'L', pdfium.FPDFBitmap_Gray
        else:
            px = 'BGR', pdfium.FPDFBitmap_BGR
    
    return px


_clformat_pil = {
    'BGRA': 'RGBA',
    'BGR': 'RGB',
    'L': 'L',
}


class BitmapDataHolder:
    """ Class to store a PDFium bitmap handle and the pointer to its data. """
    
    def __init__(self, bm_handle, bm_array_ptr):
        self.bm_handle = bm_handle
        self.bm_array_ptr = bm_array_ptr
    
    def get_data(self):
        """ Retrieve the raw bitmap data as a ctypes ubyte array. """
        return self.bm_array_ptr.contents
    
    def close(self):
        """ Release resources associated to the bitmap. """
        pdfium.FPDFBitmap_Destroy(self.bm_handle)


def render_page_base(
        pdf,
        page_index = 0,
        scale = 1,
        rotation = 0,
        colour = (255, 255, 255, 255),
        annotations = True,
        greyscale = False,
        optimise_mode = OptimiseMode.none,
        crop = (0, 0, 0, 0),
    ):
    """
    Render a single PDF page to ctypes data using PDFium.
    This is the base function for :func:`.render_pdf_tobytes` and :func:`.render_pdf_topil`.
    
    Parameters:
        
        pdf (``FPDF_DOCUMENT``):
            A raw PDFium document handle.
        
        page_index (int):
            Zero-based index of the page to render.
        
        scale (float):
            Define the quality (or size) of the image.
            By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the number of pixels that represent one point.
            Higher values increase quality, file size and rendering duration, while lower values reduce them.
            Note that UserUnit is not taken into account, so if you are using pypdfium2 in conjunction with an other PDF library, you may want to check for a possible ``/UserUnit`` in the page dictionary and multiply this scale factor with it.
        
        rotation (int):
            Rotate the page by 90, 180, or 270 degrees. Value 0 means no rotation.
        
        colour (None | typing.Tuple[int, int, int, typing.Optional[int]]):
            Page background colour. Defaults to white.
            It can either be :data:`None`, or values of red, green, blue, and alpha ranging from 0 to 255.
            If :data:`None`, the bitmap will not be filled with a colour, resulting in transparent background.
            For RGB, 0 will include nothing of the colour in question, while 255 will completely include it. For Alpha, 0 means full transparency, while 255 means no transparency.
        
        annotations (bool):
            Whether to render page annotations.
        
        greyscale (bool):
            Whether to render in greyscale mode (no colours).
        
        optimise_mode (OptimiseMode):
            Optimise rendering for LCD displays or for printing.
        
        crop (typing.Sequence[int|float]):
            Amount in PDF canvas units to cut off from page borders (left, bottom, right, top).
            Crop is applied after rotation.
    
    Returns:
    
        :class:`BitmapDataHolder`, :class:`str`, ``Tuple[int, int]`` – Bitmap data holder, used colour format, and image size.
        
        Call :meth:`BitmapDataHolder.get_data` to obtain the raw ctypes ubyte array. ``bytes(data_holder.get_data())`` may be used to acquire an independent copy of the data as Python bytes. When you have finished working with the ctypes ubyte array, call :meth:`BitmapDataHolder.close` to release allocated memory.
        
        The colour format can be ``BGRA``, ``BGR``, or ``L``, depending on the parameters *colour* and *greyscale*.
        
        The image size is given in pixels as a tuple (width, height).
    """
    
    if colour is None:
        fpdf_colour, use_alpha = None, True
    else:
        fpdf_colour, use_alpha = colour_as_hex(*colour)
    
    cl_format, cl_pdfium = _get_clformat(use_alpha, greyscale)
    n_colours = len(cl_format)
    
    form_config = pdfium.FPDF_FORMFILLINFO(2)
    form_fill = pdfium.FPDFDOC_InitFormFillEnvironment(pdf, form_config)
    page = open_page(pdf, page_index)
    
    src_width  = math.ceil(pdfium.FPDF_GetPageWidthF(page)  * scale)
    src_height = math.ceil(pdfium.FPDF_GetPageHeightF(page) * scale)
    if rotation in (90, 270):
        src_width, src_height = src_height, src_width
    
    crop = [math.ceil(c*scale) for c in crop]
    width  = src_width  - crop[0] - crop[2]
    height = src_height - crop[1] - crop[3]
    if any(d < 1 for d in (width, height)):
        raise ValueError("Crop exceeds page dimensions (in px): width %s, height %s, crop %s" % (src_width, src_height, crop))
    
    bitmap = pdfium.FPDFBitmap_CreateEx(width, height, cl_pdfium, None, width*n_colours)
    if fpdf_colour is not None:
        pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, fpdf_colour)
    
    render_flags = 0x00
    if annotations:
        render_flags |= pdfium.FPDF_ANNOT
    if greyscale:
        render_flags |= pdfium.FPDF_GRAYSCALE
    
    if optimise_mode is OptimiseMode.none:
        pass
    elif optimise_mode is OptimiseMode.lcd_display:
        render_flags |= pdfium.FPDF_LCD_TEXT
    elif optimise_mode is OptimiseMode.printing:
        render_flags |= pdfium.FPDF_PRINTING
    else:
        raise ValueError("Invalid optimise_mode %s" % optimise_mode)
    
    render_args = (bitmap, page, -crop[0], -crop[3], src_width, src_height, translate_rotation(rotation), render_flags)
    pdfium.FPDF_RenderPageBitmap(*render_args)
    pdfium.FPDF_FFLDraw(form_fill, *render_args)
    
    cbuf_ptr = pdfium.FPDFBitmap_GetBuffer(bitmap)
    cbuf_array_ptr = ctypes.cast(cbuf_ptr, ctypes.POINTER(ctypes.c_ubyte * (width*height*n_colours)))
    data_holder = BitmapDataHolder(bitmap, cbuf_array_ptr)
    
    pdfium.FPDF_ClosePage(page)
    pdfium.FPDFDOC_ExitFormFillEnvironment(form_fill)
    
    return data_holder, cl_format, (width, height)


def render_page_tobytes(*args, **kws):
    """
    Render a single page to bytes. Parameters are the same as for :func:`.render_page_base`.
    
    Returns:
         :class:`bytes`, :class:`str`, Tuple[int, int]
    """
    data_holder, cl_format, size = render_page_base(*args, **kws)
    data = bytes( data_holder.get_data() )
    data_holder.close()
    return data, cl_format, size


def render_page_topil(*args, **kws):
    """
    Render a single page to a :mod:`PIL` image. Parameters are the same as for :func:`.render_page_base`.
    
    Returns:
        :class:`PIL.Image.Image`
    """
    
    if not have_pil:
        raise RuntimeError("Pillow library needs to be installed for render_page_topil().")
    
    data_holder, cl_format, size = render_page_base(*args, **kws)
    pil_image = PIL.Image.frombytes(_clformat_pil[cl_format], size, data_holder.get_data(), "raw", cl_format, 0, 1)
    data_holder.close()
    
    return pil_image



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
from ..conftest import TestFiles
import pypdfium2 as pdfium


@pytest.fixture
def doc():
    doc = pdfium.PdfDocument(TestFiles.text)
    yield doc
    doc.close()

@pytest.fixture
def page(doc):
    page = doc.get_page(0)
    yield page
    pdfium.close_page(page)

@pytest.fixture
def textpage(page):
    textpage = pdfium.PdfTextPage(page)
    yield textpage
    textpage.close()


def test_gettext_page(textpage):
    text = textpage.get_text()
    print(text)
    assert len(text) == 438
    assert text.startswith("Lorem ipsum dolor sit amet,\r\n")
    assert text.endswith("\r\nofficia deserunt mollit anim id est laborum.")


@pytest.mark.parametrize("loose", [False, True])
def test_getcharbox(textpage, loose):
    for index in range( textpage.count_chars() ):
        box = textpage.get_charbox(index, loose=loose)
        assert all( isinstance(val, (int, float)) for val in box )
        assert box[0] <= box[2] and box[1] <= box[3]


def test_getrectboxes(textpage):
    rects = textpage.get_rectboxes()
    
    first_rect = next(rects)
    assert pytest.approx(first_rect, abs=1) == [58, 767, 258, 782]
    first_text = textpage.get_text(*first_rect)
    assert first_text == "Lorem ipsum dolor sit amet,"
    print(first_rect, first_text)
    
    i = 0
    for rect in rects:
        assert len(rect) == 4
        assert 56 < rect[0] < 59
        text = textpage.get_text(*rect)
        print(rect, text)
        assert isinstance(text, str)
        assert len(text) <= 66
        i += 1
    
    assert i == 9
    assert text == "officia deserunt mollit anim id est laborum."


def test_textpage_empty():
    doc = pdfium.PdfDocument(TestFiles.empty)
    textpage = doc.get_textpage(0)
    
    assert textpage.get_text() == ""
    assert textpage.count_chars() == 0
    assert textpage.count_rects() == 0
    assert textpage.get_index(0, 0, 0, 0) is None
    assert [r for r in textpage.get_rectboxes()] == []
    
    searcher = textpage.search("a")
    assert searcher.get_next() is None
    searcher.close()
    
    with pytest.raises(ValueError, match="Character index 0 is out of bounds. The maximum index is -1."):
        textpage.get_charbox(0)
    with pytest.raises(ValueError, match="Text length must be >0."):
        textpage.search("")
    
    textpage.close()
    doc.close()


def test_area_oob(textpage):
    areas = [
        (-1, 0, 0, 0),   # left too low
        (0, -1, 0, 0),   # bottom too low
        (0, 0, -1, 0),   # right too low
        (0, 0, 0, -1),   # top too low
        (600, 0, 0, 0),  # left too high
        (0, 900, 0, 0),  # bottom too high
        (0, 0, 600, 0),  # right too high
        (0, 0, 0, 900),  # top too high
        (100, 0, 50, 0), # left higher than right
        (0, 100, 0, 50), # bottom higher than top
    ]
    for area in areas:
        with pytest.raises(ValueError, match="Invalid page area requested."):
            textpage.get_text(*area)


def test_search_text(textpage):
    searcher = textpage.search("labor")
    assert isinstance(searcher, pdfium.TextSearcher)
    
    occ_1a = searcher.get_next()
    occ_2a = searcher.get_next()
    occ_3a = searcher.get_next()
    
    occ_4a = searcher.get_next()
    
    occ_2b = searcher.get_prev()
    occ_1b = searcher.get_prev()
    
    for occ in (occ_1a, occ_2a, occ_3a):
        for box in occ:
            assert all(isinstance(val, (int, float)) for val in box)
            assert len(box) == 4
            assert box[0] <= box[2]
            assert box[1] <= box[3]
    
    assert occ_1a == occ_1b and occ_2a == occ_2b
    assert occ_4a is None
    
    searcher.close()


def test_get_index(page, textpage):
    
    height = pdfium.FPDF_GetPageHeightF(page)
    x, y = (60, height-66)
    
    n_chars = textpage.count_chars()
    index = textpage.get_index(x, y, 5, 5)
    assert index < n_chars and index == 0
    
    charbox = textpage.get_charbox(index)
    char = textpage.get_text(*charbox)
    assert char == "L"



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-FileCopyrightText: 2020 Yinlin Hu <huyinlin@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0

import sys
import math
import ctypes
from PIL import Image
import pypdfium2 as pdfium


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print("Usage: example.py somefile.pdf")
        sys.exit()
    
    filename = sys.argv[1]
    
    doc = pdfium.FPDF_LoadDocument(filename, None)
    page_count = pdfium.FPDF_GetPageCount(doc)
    assert page_count >= 1
    
    form_config = pdfium.FPDF_FORMFILLINFO(2)
    form_fill = pdfium.FPDFDOC_InitFormFillEnvironment(doc, form_config)
    
    page = pdfium.FPDF_LoadPage(doc, 0)
    width = math.ceil(pdfium.FPDF_GetPageWidthF(page))
    height = math.ceil(pdfium.FPDF_GetPageHeightF(page))
    
    bitmap = pdfium.FPDFBitmap_Create(width, height, 0)
    pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, 0xFFFFFFFF)
    
    render_args = [bitmap, page, 0, 0, width, height, 0,  pdfium.FPDF_LCD_TEXT | pdfium.FPDF_ANNOT]
    pdfium.FPDF_RenderPageBitmap(*render_args)
    pdfium.FPDF_FFLDraw(form_fill, *render_args)
    
    cbuffer = pdfium.FPDFBitmap_GetBuffer(bitmap)
    buffer = ctypes.cast(cbuffer, ctypes.POINTER(ctypes.c_ubyte * (width * height * 4)))
    
    img = Image.frombuffer("RGBA", (width, height), buffer.contents, "raw", "BGRA", 0, 1)
    img.save("out.png")
    
    pdfium.FPDFBitmap_Destroy(bitmap)
    pdfium.FPDF_ClosePage(page)
    
    pdfium.FPDFDOC_ExitFormFillEnvironment(form_fill)
    pdfium.FPDF_CloseDocument(doc)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sysconfig
from wheel.bdist_wheel import bdist_wheel
import pytest
from pl_setup import setup_base
from pl_setup.packaging_base import PlatformNames, VerNamespace
from ..conftest import pl_names


ExpectedTags = (
    (PlatformNames.darwin_x64,    "macosx_10_11_x86_64.macosx_11_0_x86_64.macosx_12_0_x86_64"),
    (PlatformNames.darwin_arm64,  "macosx_11_0_arm64.macosx_12_0_arm64"),
    (PlatformNames.linux_x64,     "manylinux_2_17_x86_64.manylinux2014_x86_64"),
    (PlatformNames.linux_x86,     "manylinux_2_17_i686.manylinux2014_i686"),
    (PlatformNames.linux_arm64,   "manylinux_2_17_aarch64.manylinux2014_aarch64"),
    (PlatformNames.linux_arm32,   "manylinux_2_17_armv7l.manylinux2014_armv7l"),
    (PlatformNames.musllinux_x64, "musllinux_1_2_x86_64"),
    (PlatformNames.musllinux_x86, "musllinux_1_2_i686"),
    (PlatformNames.windows_x64,   "win_amd64"),
    (PlatformNames.windows_arm64, "win_arm64"),
    (PlatformNames.windows_x86,   "win32"),
    (PlatformNames.sourcebuild,   sysconfig.get_platform().replace('-','_').replace('.','_')),
)


def test_expected_tags():
    assert len(pl_names) == len(ExpectedTags)
    for platform, tag in ExpectedTags:
        assert hasattr(PlatformNames, platform)
        assert isinstance(tag, str)


def test_get_tag():
    for platform, tag in ExpectedTags:
        assert setup_base._get_tag(platform) == tag

def test_unknown_tag():
    plat_dir = "win_amd74"
    with pytest.raises(ValueError, match="Unknown platform directory %s" % plat_dir):
        setup_base._get_tag(plat_dir)

def test_get_bdist():
    for platform, tag in ExpectedTags:
        bdist_cls = setup_base._get_bdist(platform)
        assert issubclass(bdist_cls, bdist_wheel)

def test_setupkws():
    assert "version" in setup_base.SetupKws
    assert setup_base.SetupKws["version"] == VerNamespace["V_PYPDFIUM2"]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause  

import io
from PIL import Image
from ..conftest import (
    TestFiles,
    iterate_testfiles,
)
import pypdfium2 as pdfium


def _pdfdoc_load(input_obj):
    doc = pdfium.PdfDocument(input_obj)
    page = pdfium.FPDF_LoadPage(doc.raw, 0)
    pdfium.FPDF_ClosePage(page)
    doc.close()


def test_pdfdoc_loadfiles():
    for filepath in iterate_testfiles():
        _pdfdoc_load(filepath)


def test_pdfdoc_renderpage():
    doc = pdfium.PdfDocument(TestFiles.render)
    image = doc.render_page_topil(0)
    assert isinstance(image, Image.Image)
    doc.close()


def test_pdfdoc_renderpdf():
    
    doc = pdfium.PdfDocument(TestFiles.multipage)
    
    i = 0
    for image, suffix in doc.render_pdf_topil():
        assert isinstance(image, Image.Image)
        assert isinstance(suffix, str)
        i+= 1
    assert i == 3
    
    doc.close()


def test_pdfdoc_save():
    doc = pdfium.PdfDocument(TestFiles.multipage)
    pdfium.FPDFPage_Delete(doc.raw, 0)
    buffer = io.BytesIO()
    doc.save(buffer)
    doc.close()
    assert buffer.tell() > 100000
    buffer.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
import pypdfium2 as pdfium
from ..conftest import TestFiles


def test_boxes():
    
    with pdfium.PdfContext(TestFiles.boxes) as pdf:
        
        page_a = pdfium.open_page(pdf, 0)
        page_b = pdfium.open_page(pdf, 1)
        
        test_cases = [
            (page_a, pdfium.get_mediabox, (0,  0,  612, 792)),
            (page_b, pdfium.get_mediabox, (0,  0,  595, 842)),
            (page_b, pdfium.get_cropbox,  (10, 10, 585, 832)),
            (page_b, pdfium.get_bleedbox, (20, 20, 575, 822)),
            (page_b, pdfium.get_trimbox,  (30, 30, 565, 812)),
            (page_b, pdfium.get_artbox,   (40, 40, 555, 802)),
        ]
        
        for page, box_func, exp_box in test_cases:
            assert pytest.approx( box_func(page) ) == exp_box
        
        for page in (page_a, page_b):
            pdfium.FPDF_ClosePage(page)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import atexit
import logging
from pypdfium2._namespace import *

logger = logging.getLogger(__name__)

FPDF_InitLibrary()
atexit.register(FPDF_DestroyLibrary)
